LIBRARY libgr3.dll

EXPORTS
	gr3_init
	gr3_terminate
	gr3_getrenderpathstring
	gr3_clear
	gr3_drawimage
	gr3_getimage
	gr3_export
	gr3_createmesh
	gr3_createmesh_nocopy
	gr3_createindexedmesh
	gr3_createindexedmesh_nocopy
	gr3_drawmesh
	gr3_deletemesh
	gr3_cameralookat
	gr3_setcameraprojectionparameters
	gr3_getcameraprojectionparameters
	gr3_setlightdirection
	gr3_setbackgroundcolor
	gr3_geterrorstring
	gr3_setlogcallback
	gr3_drawconemesh
	gr3_drawcylindermesh
	gr3_drawspheremesh
	gr3_drawcubemesh
	gr3_createisosurfacemesh
	gr3_setobjectid
	gr3_selectid
	gr3_triangulate
	gr3_triangulateindexed
	gr3_free
	gr3_getviewmatrix
	gr3_setviewmatrix
	gr3_getprojectiontype
	gr3_setprojectiontype
	gr3_createsurfacemesh
	gr3_drawmesh_grlike
	gr3_drawsurface
	gr3_surface

