LIBRARY libgr.dll

EXPORTS
	gr_activatews
	gr_adjustrange
	gr_axes
	gr_axes3d
	gr_begingraphics
	gr_beginprint
	gr_beginprintext
	gr_beginselection
	gr_cellarray
	gr_clearws
	gr_closegks
	gr_closeseg
	gr_closews
	gr_colormap
	gr_contour
	gr_copysegws
	gr_createseg
	gr_deactivatews
	gr_drawarrow
	gr_drawimage
	gr_emergencyclosegks
	gr_endgraphics
	gr_endprint
	gr_endselection
	gr_fillarea
	gr_grid
	gr_gridit
	gr_herrorbars
	gr_importgraphics
	gr_inqcolor
	gr_inqcolorfromrgb
	gr_inqcolormap
	gr_inqdspsize
	gr_inqscale
	gr_inqspace
	gr_inqtext
	gr_inqtextext
	gr_mathtex
	gr_moveselection
	gr_ndctowc
	gr_opengks
	gr_openws
	gr_polyline
	gr_polyline3d
	gr_polymarker
	gr_readimage
	gr_redrawsegws
	gr_resizeselection
	gr_selntran
	gr_setarrowstyle
	gr_setcharexpan
	gr_setcharheight
	gr_setcharspace
	gr_setcharup
	gr_setclip
	gr_setcolormap
	gr_setcolorrep
	gr_setfillcolorind
	gr_setfillintstyle
	gr_setfillstyle
	gr_setlinecolorind
	gr_setlinetype
	gr_setlinewidth
	gr_setmarkercolorind
	gr_setmarkersize
	gr_setmarkertype
	gr_setscale
	gr_setsegtran
	gr_setspace
	gr_settextalign
	gr_settextcolorind
	gr_settextfontprec
	gr_settextpath
	gr_setviewport
	gr_setwindow
	gr_setwsviewport
	gr_setwswindow
	gr_spline
	gr_surface
	gr_text
	gr_textext
	gr_tick
	gr_titles3d
	gr_updategks
	gr_updatews
	gr_verrorbars
	gr_wctondc
