MAX_TNR = 9
MWIDTH = 0.254
MHEIGHT = 0.1905
WIDTH = 1024
HEIGHT = 768
GKS_K_INTSTYLE_HOLLOW = 0
GKS_K_INTSTYLE_SOLID = 1
GKS_K_INTSTYLE_PATTERN = 2
GKS_K_INTSTYLE_HATCH = 3
PATTERNS = 120
HATCH_STYLE = 108
GKS_K_TEXT_PRECISION_STRING = 0
MAX_COLOR = 1256
GKS_K_TEXT_VALIGN_NORMAL = 0
GKS_K_TEXT_VALIGN_TOP = 1
GKS_K_TEXT_VALIGN_CAP = 2
GKS_K_TEXT_VALIGN_HALF = 3
GKS_K_TEXT_VALIGN_BASE = 4
GKS_K_TEXT_VALIGN_BOTTOM = 5
GKS_K_TEXT_HALIGN_NORMAL = 0
GKS_K_TEXT_HALIGN_LEFT = 1
GKS_K_TEXT_HALIGN_CENTER = 2
GKS_K_TEXT_HALIGN_RIGHT = 3
GKS_K_NOCLIP = 0
GKS_K_CLIP = 1
