LIBRARY libgks.dll

EXPORTS
	gks_open_gks
	gks_close_gks
	gks_open_ws
	gks_close_ws
	gks_activate_ws
	gks_deactivate_ws
	gks_clear_ws
	gks_redraw_seg_on_ws
	gks_update_ws
	gks_set_deferral_state
	gks_escape
	gks_message
	gks_polyline
	gks_polymarker
	gks_text
	gks_fillarea
	gks_cellarray
	gks_set_pline_index
	gks_set_pline_linetype
	gks_set_pline_linewidth
	gks_set_pline_color_index
	gks_set_pmark_index
	gks_set_pmark_type
	gks_set_pmark_size
	gks_set_pmark_color_index
	gks_set_text_index
	gks_set_text_fontprec
	gks_set_text_expfac
	gks_set_text_spacing
	gks_set_text_color_index
	gks_set_text_height
	gks_set_text_upvec
	gks_set_text_path
	gks_set_text_align
	gks_set_fill_index
	gks_set_fill_int_style
	gks_set_fill_style_index
	gks_set_fill_color_index
	gks_set_asf
	gks_set_color_rep
	gks_set_window
	gks_set_viewport
	gks_select_xform
	gks_set_clipping
	gks_set_ws_window
	gks_set_ws_viewport
	gks_create_seg
	gks_close_seg
	gks_delete_seg
	gks_assoc_seg_with_ws
	gks_copy_seg_to_ws
	gks_set_seg_xform
	gks_initialize_locator
	gks_request_locator
	gks_request_stroke
	gks_request_choice
	gks_request_string
	gks_read_item
	gks_get_item
	gks_interpret_item
	gks_eval_xform_matrix
	gks_inq_operating_state
	gks_inq_level
	gks_inq_wstype
	gks_inq_max_xform
	gks_inq_open_ws
	gks_inq_active_ws
	gks_inq_segn_ws
	gks_inq_color_rep
	gks_inq_pline_linetype
	gks_inq_pline_linewidth
	gks_inq_pline_color_index
	gks_inq_pmark_type
	gks_inq_pmark_size
	gks_inq_pmark_color_index
	gks_inq_text_fontprec
	gks_inq_text_expfac
	gks_inq_text_spacing
	gks_inq_text_color_index
	gks_inq_text_height
	gks_inq_text_upvec
	gks_inq_text_path
	gks_inq_text_align
	gks_inq_fill_int_style
	gks_inq_fill_style_index
	gks_inq_fill_color_index
	gks_inq_open_segn
	gks_inq_current_xformno
	gks_inq_xform
	gks_inq_clip
	gks_inq_ws_conntype
	gks_inq_ws_category
	gks_inq_text_extent
	gks_inq_max_ds_size
	gks_emergency_close
	gks_set_text_slant
	gks_draw_image
	gks_set_shadow
	gks_set_transparency
	gks_set_coord_xform
	gks_begin_selection
	gks_end_selection
	gks_move_selection
	gks_resize_selection
	gks_inq_bbox
	gks_inq_text_slant
	gks_a_error_info
	gks_errno
	GOPKS
	GCLKS
	GOPWK
	GCLWK
	GACWK
	GDAWK
	GCLRWK
	GRSGWK
	GUWK
	GSDS
	GESC
	GMSG
	GPL
	GPM
	GTX
	GTXS
	GFA
	GCA
	GSPLI
	GSLN
	GSLWSC
	GSPLCI
	GSPMI
	GSMK
	GSMKSC
	GSPMCI
	GSTXI
	GSTXFP
	GSCHXP
	GSCHSP
	GSTXCI
	GSCHH
	GSCHUP
	GSTXP
	GSTXAL
	GSFAI
	GSFAIS
	GSFASI
	GSFACI
	GSASF
	GSCR
	GSWN
	GSVP
	GSELNT
	GSCLIP
	GSWKWN
	GSWKVP
	GCRSG
	GCLSG
	GDSG
	GASGWK
	GCSGWK
	GSSGT
	GINLC
	GRQLC
	GRQSK
	GRQCH
	GRQST
	GRDITM
	GGTITM
	GIITM
	GEVTM
	GQOPS
	GQLVKS
	GQEWK
	GQMNTN
	GQOPWK
	GQACWK
	GQSGWK
	GQLN
	GQLWSC
	GQPLCI
	GQMK
	GQMKSC
	GQPMCI
	GQTXFP
	GQCHXP
	GQCHSP
	GQTXCI
	GQCHH
	GQCHUP
	GQTXP
	GQTXAL
	GQFAIS
	GQFASI
	GQFACI
	GQOPSG
	GQCNTN
	GQNT
	GQCLIP
	GQWKC
	GQWKCA
	GQTXX
	GQTXXS
	GQDSP
	GECLKS
